//------------------------------------------------------------------------------  
//title: VMEBus Library for Universe Driver
//version: Linux 1.1
//date: April 1999
//designer: Michael Wyrick                                                      
//programmer: Michael Wyrick                                                    
//company: Umbra System Inc.
//platform: Linux 2.2.x, 2.4.x                                    
//language: GCC 2.95, 3.0
//module: 
//------------------------------------------------------------------------------  
//  Purpose:                                                                    
//  Docs:                                                                       
//------------------------------------------------------------------------------  
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <string.h>

#include "universe.h"
#include "libvme-c.h"

#ifdef __KERNEL__
#include <linux/kernel.h>
#define	MALLOC	kmalloc
#define FREE	kfree
#else
#include <stdlib.h>
#define MALLOC	malloc
#define FREE	free
#endif

static struct VME_Exception exception = {0,0,0,0};

static void throw( int nAddress, int nCount, int nType, int nException )
{
	exception.nAddress = nAddress;
	exception.nCount = nCount;
	exception.nAccessType = nType;
	exception.nException = nException;
}

//#define DEBUG

//-----------------------------------------------------------------------------
// Function   : VME_SwapEndian
// Inputs     : data, data length
// Outputs    : 
// Description: Flip the Endianess of a buffer
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
void VME_SwapEndian( void* pData, int nBytes )
{
	int i;
	char buffer[16];
	memset( buffer, 0, 16 );

	if ( nBytes > 16 )
		nBytes = 16;
	
	for ( i = 0; i < nBytes; i++ )
	{
		buffer[nBytes-1-i] = ((char*)pData)[i];
	}
	for ( i = 0; i < nBytes; i++ )
	{
		((char*)pData)[i] = buffer[i];
	}
}

//-----------------------------------------------------------------------------
// Function   : VMEBus construstor
// Inputs     :
//   addr      = VME Base Address to map
//   count     = Size of mapping window
//   space     = which VME address space the memory is in (i.e. user/data etc.)
//   which_img = which of the four universe images to use
// Outputs    :
// Description:
//   Add a space description here
//   if which_img is -1 then scan for and pick the next free image,
//   only image 3 can be used for SUPERVISOR mappings
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
struct VME_Bus* VME_Create( unsigned int addr, int count, char space, int which_img )
{
	struct VME_Bus*	pVME;
	int				to   = 0;
	int				am   = 0;
	int				size = 0;
	
	pVME = (struct VME_Bus*)MALLOC( sizeof(struct VME_Bus) );
	if ( !pVME )
		return NULL;
	
	pVME->fSwapEndian = 1;
	pVME->nCTL = 0x00800000;
	
	if (space & VME_SUPERVISOR)     // Only image can use supervisor space
		which_img = 3;
	
	if (which_img == -1) 
	{  // Auto Image
		if ((pVME->hVME = open("//dev//vme_m0",O_RDWR,0)) > 0) 
		{
			pVME->uPCIBaseAddress  = 0xC0000000;
		}
		else 
		{
			if ((pVME->hVME = open("//dev//vme_m1",O_RDWR,0)) > 0) 
			{
				pVME->uPCIBaseAddress  = 0xC4000000;
			}
			else 
			{
				if ((pVME->hVME = open("//dev//vme_m2",O_RDWR,0)) > 0) 
				{
					pVME->uPCIBaseAddress  = 0xC8000000;
				}
				else 
				{
					if ((pVME->hVME = open("//dev//vme_m3",O_RDWR,0)) > 0) 
					{
						pVME->uPCIBaseAddress  = 0xCD000000;
					}
				}
			}
		}
	}
	else 
	{ 
		switch (which_img) 
		{  // Force Selected image
			case 0 :
				pVME->hVME = open("//dev//vme_m0",O_RDWR,0);
				pVME->uPCIBaseAddress  = 0xC0000000;
				break;

			case 1 :
				pVME->hVME = open("//dev//vme_m1",O_RDWR,0);
				pVME->uPCIBaseAddress  = 0xC4000000;
				break;

			case 2 :
				pVME->hVME = open("//dev//vme_m2",O_RDWR,0);
				pVME->uPCIBaseAddress  = 0xC8000000;
				break;

			case 3 :
				pVME->hVME = open("//dev//vme_m3",O_RDWR,0);
				pVME->uPCIBaseAddress  = 0xCC000000;
				break;
		}
	}
	pVME->hUniverse = open("//dev//vme_ctl",O_RDWR,0);
  
	if (pVME->hVME < 0 || pVME->hUniverse < 0) 
	{
		if ( pVME->hVME < 0 )
			close( pVME->hVME );
		FREE( pVME );
		throw( 0,0,0, VME_EXCEPTION_BUSY );
		return NULL;
	}
  
	size = space & VME_SIZE;
  
	switch (size) 
	{
		case VME_SIZE_DO8:
			pVME->nCTL &= ~CTL_VDW;
			pVME->nCTL |= CTL_VDW_8;
			break;  

		case VME_SIZE_D16:
			pVME->nCTL &= ~CTL_VDW;
			pVME->nCTL |= CTL_VDW_16;
			break;  

		case VME_SIZE_D32:
			pVME->nCTL &= ~CTL_VDW;
			pVME->nCTL |= CTL_VDW_32;
			break;  

		case VME_SIZE_D64:
			pVME->nCTL &= ~CTL_VDW;
			pVME->nCTL |= CTL_VDW_64;
			break;  
	}
  
	am = space & 0x03;
	     
	if (space & VME_PRG_DATA) 
		pVME->nCTL |= CTL_PGM;
	else 
		pVME->nCTL &= ~CTL_PGM;
  
	if (space & VME_SUP_USR) 
		pVME->nCTL |= CTL_SUPER;
	else 
		pVME->nCTL &= ~CTL_SUPER;

	if (space & VME_CYCLE) 
		pVME->nCTL |= CTL_VCT;
	else 
		pVME->nCTL &= ~CTL_VCT;

	if (am == VME_A16) 
	{
		pVME->nVMESpace = am;
		pVME->nCTL &= ~CTL_VAS;
		pVME->nCTL |= CTL_VAS_A16;
	} 
	else if (am == VME_A24) 
	{
		pVME->nVMESpace = am;
		pVME->nCTL &= ~CTL_VAS;
		pVME->nCTL |= CTL_VAS_A24;
	}
	else if (am == VME_A32) 
	{
		pVME->nVMESpace = am;
		pVME->nCTL &= ~CTL_VAS;
		pVME->nCTL |= CTL_VAS_A32;
	}
	else if (am == VME_CR_CSR) 
	{
		pVME->nVMESpace = am;
		pVME->nCTL &= ~CTL_VAS;
		pVME->nCTL |= CTL_VAS_CR_CSR;
	}
	else 
	{
		throw( 0,0,0, VME_EXCEPTION_BADMAP );
	}

	pVME->uPCIBoundAddress = pVME->uPCIBaseAddress + (((count-1) / 0x10000)+1)*0x10000;
	pVME->uVMEBaseAddress  = addr;
	to = -pVME->uPCIBaseAddress + pVME->uVMEBaseAddress;
	
	pVME->nCTL &= ~CTL_EN;    // Disable Slave
	ioctl( pVME->hVME, IOCTL_SET_CTL, pVME->nCTL );
	ioctl( pVME->hVME, IOCTL_SET_TO, to );
	ioctl( pVME->hVME, IOCTL_SET_BD, pVME->uPCIBoundAddress );
	ioctl( pVME->hVME, IOCTL_SET_BS, pVME->uPCIBaseAddress );  // Must set this after BD
	pVME->nCTL |= CTL_EN;     // Enable Slave
	ioctl( pVME->hVME, IOCTL_SET_CTL, pVME->nCTL );
	
	if ( space & VME_DMA_MODE )
		ioctl( pVME->hVME, IOCTL_SET_MODE, MODE_DMA );
	else
		ioctl( pVME->hVME, IOCTL_SET_MODE, MODE_PROGRAMMED );
	
	return pVME;
}

//-----------------------------------------------------------------------------
// Function   : ~VMEBus
// Inputs     : 
// Outputs    :
// Description: Close handles to Universe Driver
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
void VME_Destroy( struct VME_Bus* pVME )
{
	close( pVME->hVME );
	close( pVME->hUniverse );
	FREE( pVME );
}

//-----------------------------------------------------------------------------
// Function   : ReadUniReg
// Inputs     : 
//   int reg - Universe Register address to read
// Outputs    : value of universe register
// Description:
// Remarks    : See Universe ca91c042 documentation for registers
// History    :
//-----------------------------------------------------------------------------
unsigned int VME_ReadUniReg( struct VME_Bus* pVME, int nRegister )
{
	unsigned int v;
	
	lseek( pVME->hUniverse, nRegister, SEEK_SET );
	read( pVME->hUniverse, &v, 4 );

	return v;
}

//-----------------------------------------------------------------------------
// Function   : WriteUniReg
// Inputs     : 
//   int reg - Universe Register address to read
//   uint v  - value to write to the register
// Outputs    :
// Description: Write a value to the universe register
// Remarks    : See Universe ca91c042 documentation for registers
// History    :
//-----------------------------------------------------------------------------
void VME_WriteUniReg( struct VME_Bus* pVME, int nRegister, unsigned int uData )
{
	lseek( pVME->hUniverse, nRegister, SEEK_SET );
	write( pVME->hUniverse, &uData, 4 );
}

//-----------------------------------------------------------------------------
// Function   : ReadByte
// Inputs     : 
//   int addr - VME byte Address to Read the byte from
// Outputs    : the byte that was read
// Description: Read a byte from the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//-----------------------------------------------------------------------------
unsigned char VME_ReadByte( struct VME_Bus* pVME, int nRegister )
{  
	int c;
	unsigned char v[4];
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = read( pVME->hVME, v, 1 );

#ifdef DEBUG
	printf("<libvme:ReadByte> addr=%08X v=%02X s=%01X\n", nRegister, *v, c);
#endif

	if (c != 1) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_READ, VME_EXCEPTION_BUSERROR );
	}
	else
		pVME->nLastError = 0;

	return v[0];
}

//-----------------------------------------------------------------------------
// Function   : ReadWord
// Inputs     : 
//   int addr - VME byte Address to Read the word from
// Outputs    : the value of the word
// Description: Read a word from the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//-----------------------------------------------------------------------------
unsigned short VME_ReadWord( struct VME_Bus* pVME, int nRegister )
{      
	int c;
	unsigned short v[2];	 
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = read( pVME->hVME, &v, 2 );
	
#ifdef DEBUG
	printf("<libvme:ReadWord> addr=%08X v=%02X s=%01X\n", nRegister, *v, c);
#endif
	
	if (c != 2) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_READ, VME_EXCEPTION_BUSERROR );
	} 
	else
		pVME->nLastError = 0;
	
	if ( pVME->fSwapEndian )
		VME_SwapEndian( &v[0], sizeof(unsigned short) );

	return v[0]; 
}

//-----------------------------------------------------------------------------
// Function   : ReadLong
// Inputs     : 
//   int addr - VME byte Address to Read the long word from
// Outputs    : the value of the long word
// Description: Read a long from the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//-----------------------------------------------------------------------------
unsigned int VME_ReadLong( struct VME_Bus* pVME, int nRegister )
{    
	int c;
	unsigned int v;
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = read( pVME->hVME, &v, 4 );

#ifdef DEBUG
	printf("<libvme:Readlong> addr=%08X v=%02X s=%01X\n", nRegister, v, c);
#endif

	if (c != 4) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_READ, VME_EXCEPTION_BUSERROR );
	}
	else
		pVME->nLastError = 0;
	
	if ( pVME->fSwapEndian )
		VME_SwapEndian( &v, sizeof(unsigned int) );

	return v; 
}

//-----------------------------------------------------------------------------
// Function   : WriteByte
// Inputs     : 
//   int addr - VME byte Address to write the byte to
//   uchar v  - the byte to write
// Outputs    :
// Description: Write a Byte to the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//-----------------------------------------------------------------------------
void VME_WriteByte( struct VME_Bus* pVME, int nRegister, unsigned char ucData )
{     
	int c;
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = write( pVME->hVME, &ucData, 1 );

#ifdef DEBUG
	printf("<libvme:WriteByte> addr=%08X v=%02X s=%01X\n", nRegister, ucData, c);
#endif

	if (c != 1) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_WRITE, VME_EXCEPTION_BUSERROR );
	} 
	else
		pVME->nLastError = 0;
}

//-----------------------------------------------------------------------------
// Function   : WriteWord
// Inputs     : 
//   int addr - VME byte Address to write the word to
//   ushort v - the word to write
// Outputs    :
// Description: Write a word to the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//-----------------------------------------------------------------------------
void VME_WriteWord( struct VME_Bus* pVME, int nRegister, unsigned short usData )
{     
	int c;
	unsigned short result; 
	
	if ( pVME->fSwapEndian )
		VME_SwapEndian( &usData, sizeof(unsigned short) );
	result = usData; 
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = write( pVME->hVME, &result, 2 );

#ifdef DEBUG
	printf("<libvme:WriteWord> addr=%08X v=%02X s=%01X\n", nRegister, result, c);
#endif

	if (c != 2) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_WRITE, VME_EXCEPTION_BUSERROR );
	} 
	else
		pVME->nLastError = 0;
}

//-----------------------------------------------------------------------------
// Function   : WriteLong
// Inputs     : 
//   int addr - VME byte Address to write the long word to
//   uint v   - the long word to write
// Outputs    :
// Description: Write a long word to the VME Bus
// Remarks    : Will throw a BusError exception if there is an error
// History    :
//  99/Apr/9 Fixed unsigned short result to unsigned long result
//           This would not write a long correctly before this fix.
//----------------------------------------------------------------------------
void VME_WriteLong( struct VME_Bus* pVME, int nRegister, unsigned int uData )
{		     
	int c;
	unsigned long result; 
	
	if ( pVME->fSwapEndian )
		VME_SwapEndian( &uData, sizeof(unsigned int) );
	result = uData; 
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = write( pVME->hVME, &result, 4 );

#ifdef DEBUG
	printf("<libvme:WriteLong> addr=%08X v=%08lX s=%01X\n", nRegister, result, c);
#endif

	if (c != 4) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_WRITE, VME_EXCEPTION_BUSERROR );
	}
	else
		pVME->nLastError = 0;
}

//-----------------------------------------------------------------------------
// Function   : ReadWord_wa
// Inputs     :
//   int addr - the Word Address of the word to read
// Outputs    : the word that was read from the VME bus
// Description: This function uses word addressing insted of byte addressing
// Remarks    :
//   This function is used when the native register map for a board is in words
// History    :
//-----------------------------------------------------------------------------
unsigned short VME_ReadWord_wa( struct VME_Bus* pVME, int nRegister )
{      
	return VME_ReadWord( pVME, nRegister * 2 );
}

//-----------------------------------------------------------------------------
// Function   : WriteWord_wa
// Inputs     :
//   int addr - the Word Address of the word to write
//   ushort v - the word to write
// Outputs    : 
// Description: This function uses word addressing insted of byte addressing
// Remarks    :
//   This function is used when the native register map for a board is in words
// History    :
//-----------------------------------------------------------------------------
void VME_WriteWord_wa( struct VME_Bus* pVME, int nRegister, unsigned short usData )
{      
	VME_WriteWord( pVME, nRegister * 2, usData );
}

//-----------------------------------------------------------------------------
// Function   : ReadLong_la
// Inputs     :
//   int addr - the long word  Address of the long word to read
// Outputs    : the long word that was read from the VME bus
// Description: This function uses long word addressing insted of byte 
//              addressing
// Remarks    :
//   This function is used when the native register map for a board is in 
//   long words
// History    :
//-----------------------------------------------------------------------------
unsigned long VME_ReadLong_la( struct VME_Bus* pVME, int nRegister )
{      
	return VME_ReadLong( pVME, nRegister * 4 );
}

//-----------------------------------------------------------------------------
// Function   : WriteLong_la
// Inputs     :
//   int addr - the long word Address of the long word to write to
//   uint v   - the long word to write
// Outputs    : 
// Description: This function uses long word addressing insted of byte 
//              addressing
// Remarks    :
//   This function is used when the native register map for a board is in 
//   long words
// History    :
//-----------------------------------------------------------------------------
void VME_WriteLong_la( struct VME_Bus* pVME, int nRegister, unsigned long ulData )
{      
	VME_WriteLong( pVME, nRegister * 4, ulData );
}

//-----------------------------------------------------------------------------
// Function   : ReadBlock
// Inputs     :
//   int addr   - VME Byte address to read from
//   int count  - The number of BYTES to read
//   uchar*buff - Pointer to place data into (the user must provide enough room)
// Outputs    : The number of Byte read
// Description:
// Remarks    : Will throw a BusError exception if there is an error. This 
//   function can be used to do DMA transfers if DMA mode is enabled.
// History    :
//-----------------------------------------------------------------------------
int VME_ReadBlock( struct VME_Bus* pVME, int nRegister, int nCount, unsigned char* pBuff )
{  
	int c;
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = read( pVME->hVME, pBuff, nCount );
	if (c != nCount) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_READ, VME_EXCEPTION_BUSERROR );
	} 
	else
		pVME->nLastError = 0;

	return c;
}

//-----------------------------------------------------------------------------
// Function   : WriteBlock
// Inputs     :
//   int addr   - VME Byte address to read from
//   int count  - The number of BYTES to write
//   uchar*buff - Pointer to data 
// Outputs    : The number of Byte written
// Description:
// Remarks    : Will throw a BusError exception if there is an error. This 
//   function can be used to do DMA transfers if DMA mode is enabled.
// History    :
//-----------------------------------------------------------------------------
int VME_WriteBlock( struct VME_Bus* pVME, int nRegister, int nCount, unsigned char* pBuff )
{  
	int c;
	
	lseek( pVME->hVME, nRegister, SEEK_SET );
	c = write( pVME->hVME, pBuff, nCount );
	if (c != nCount) 
	{
		pVME->nLastError = -1;
		throw( nRegister + c, c, VME_TYPE_WRITE, VME_EXCEPTION_BUSERROR );
	} 
	else
		pVME->nLastError = 0;

	return c;
}

//-----------------------------------------------------------------------------
// Function   : SwapEndianOff
// Inputs     :
// Outputs    : returns the value of iSwapEndian
// Description: Use this function to disable the Endian Swapping that this
//   library will do on word and long word accesses
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
int VME_SwapEndianOff( struct VME_Bus* pVME )
{
	pVME->fSwapEndian = 0;
	return (pVME->fSwapEndian);
}

//-----------------------------------------------------------------------------
// Function   : SwapEndianOn
// Inputs     :
// Outputs    : returns the value of iSwapEndian
// Description: Use this function to enable the Endian Swapping that this
//   library will do on word and long word accesses
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
int VME_SwapEndianOn( struct VME_Bus* pVME )
{
	pVME->fSwapEndian = 1;
	return (pVME->fSwapEndian);
}

//-----------------------------------------------------------------------------
// Function   : Mode_DMA
// Inputs     :
// Outputs    :
// Description: Turn on DMA mode transfers
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
void VME_Mode_DMA( struct VME_Bus* pVME )
{
	ioctl( pVME->hVME, IOCTL_SET_MODE, MODE_DMA );
}

//-----------------------------------------------------------------------------
// Function   : Mode_Programmed
// Inputs     :
// Outputs    :
// Description: Turn off DMA mode transfers
// Remarks    :
// History    :
//-----------------------------------------------------------------------------
void VME_Mode_Programmed( struct VME_Bus* pVME )
{
	ioctl( pVME->hVME, IOCTL_SET_MODE, MODE_PROGRAMMED );
}
