//------------------------------------------------------------------------------  
//title: VMEBus Library for Universe Driver
//version: Linux 1.1
//date: April 1999
//designer: Michael Wyrick                                                      
//programmer: Michael Wyrick                                                    
//company: Umbra System Inc.
//platform: Linux 2.2.x, 2.4.x                                    
//language: GCC 2.95, 3.0
//module: 
//------------------------------------------------------------------------------  
//  Purpose:                                                                    
//  Docs:                                                                       
//------------------------------------------------------------------------------  
#ifndef VMELIB_C_H
#define VMELIB_C_H

#include "libvme.h"

//----------------------------------------------------------------------------
// VMEException Struct
//----------------------------------------------------------------------------
const unsigned char VME_TYPE_READ  = 1; 
const unsigned char VME_TYPE_WRITE = 2;
const unsigned char VME_EXCEPTION_BUSERROR	= 1;
const unsigned char VME_EXCEPTION_BUSY		= 2;
const unsigned char VME_EXCEPTION_BADMAP	= 3;

struct VME_Exception
{
	int		nAddress;
	int		nCount;
	int		nAccessType;
	int		nException;
};

//----------------------------------------------------------------------------
// VMEBus Struct
//----------------------------------------------------------------------------
struct VME_Bus
{
	int		fSwapEndian;
	int		hVME;
	int		hUniverse;
	int		nVMESpace;
	int		nCTL;
	
	unsigned int	uPCIBaseAddress;
	unsigned int	uPCIBoundAddress;
	unsigned int	uVMEBaseAddress;
	
	int		nLastError;
};

struct VME_Bus*	VME_Create( unsigned int addr, int count, char space, int image );
void			VME_Destroy( struct VME_Bus* pStruct );

unsigned int	VME_ReadUniReg( struct VME_Bus* pStruct, int nRegister );
void			VME_WriteUniReg( struct VME_Bus* pStruct, int nRegister, unsigned int uData );

unsigned char	VME_ReadByte( struct VME_Bus* pStruct, int nRegister );
unsigned short	VME_ReadWord( struct VME_Bus* pStruct, int nRegister );
unsigned int	VME_ReadLong( struct VME_Bus* pStruct, int nRegister );

void			VME_WriteByte( struct VME_Bus* pStruct, int nRegister, unsigned char ucData );
void			VME_WriteWord( struct VME_Bus* pStruct, int nRegister, unsigned short usData );
void			VME_WriteLong( struct VME_Bus* pStruct, int nRegister, unsigned int uData );

unsigned short	VME_ReadWord_wa( struct VME_Bus* pStruct, int nRegister );  
void			VME_WriteWord_wa( struct VME_Bus* pStruct, int nRegister, unsigned short usData );

unsigned long	VME_ReadLong_la( struct VME_Bus* pStruct, int nRegister );  
void			VME_WriteLong_la( struct VME_Bus* pStruct, int nRegister, unsigned long ulData );

int				VME_ReadBlock( struct VME_Bus* pStruct, int nRegister, int nCount, unsigned char* pBuffer );
int				VME_WriteBlock( struct VME_Bus* pStruct, int nRegister, int nCount, unsigned char* pBuffer );

int				VME_SwapEndianOn( struct VME_Bus* pStruct );
int				VME_SwapEndianOff( struct VME_Bus* pStruct );

void			VME_Mode_DMA( struct VME_Bus* pStruct );
void			VME_Mode_Programmed( struct VME_Bus* pStruct );

#endif // VMELIB_C_H
