/*
 *  header file for driver 'oms-pcixdrv.c'.  Does not present a user-space API.
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 */
#ifndef __PCIXDRV_H_
#define __PCIXDRV_H_

#include "oms-pcix.h"

#ifdef __KERNEL__

#define	INTERACTIVE_BUFFER_LEN	1024

// card specific data included in struct oms_card
struct hw_struct {
	// ******************************************************
	// physical addresses and mapped pointers into the card

	// physaddr (from pci config header BAR) of 8 byte registers
	unsigned long   regbase_physaddr;
	// mapped address of 8 byte registers
	unsigned long   regbase_addr;    // note that this is not a pointer

	// pointers to relevant registers within 8 byte space
	volatile u8*    reg_control;		// Control register
	volatile u8*    reg_status;			// Status register
	volatile u8*    reg_done;				// Done register
	volatile u8*    reg_data;				// Data register

	// physaddr (from pci config header BAR) of 8K shared memory
	unsigned long   sharedmem_physaddr;
	// mapped address of shared memory
	u8*             sharedmem;			// note that this IS a pointer
	volatile u16*		cardnum;				// physical card ID
	volatile u16*		mailbox;				// position data sampling
	volatile u16*		vec_opcode;			// vector mode type
	volatile u16*		vec_insert;			// vector queue insertion pointer
	volatile u16*		vec_process;		// vector queue process pointer

	// ******************************************************
	u8              status_flags;		// driver copy of status flags
	u8              done_flags;			// driver copy of done flags
};

// simplify access and expressions
#define	card_hw	((struct hw_struct*)(card->hardware))


// initial interrupt mask. do not enable TBE until the xmit buffer is non-empty
#define INTERRUPT_MASK (IRQ_BIT|IBF_BIT|DONE_BIT|OVRT_BIT|SLIP_BIT|CMDERR_BIT)


// Hardware specific register offsets
#define OMS_REG_CONTROL         0x00000000
#define OMS_REG_STATUS          0x00000002
#define OMS_REG_DONE            0x00000004
#define OMS_REG_DATA            0x00000006


extern struct tasklet_struct oms_tasklet;


#endif  // __KERNEL__
#endif  // __PCIXDRV_H_


